"""Utility functions for tests."""
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, List, Literal, Tuple


TEST_DATA = {
    "dataresource": {
        "content": {
            "firstName": "Joe",
            "lastName": "Jackson",
            "gender": "male",
            "age": 28,
            "address": {
                "streetAddress": "101",
                "city": "San Diego",
                "state": "CA",
            },
            "phoneNumbers": [{"type": "home", "number": "7349282382"}],
        }
    },
    "filter": {"sqlquery": "DROP TABLE myTable;"},
    "function": {},
    "mapping": {
        "prefixes": {
            "map": "http://example.org/0.0.1/mapping_ontology#",
            "onto": "http://example.org/0.2.1/ontology#",
        },
        "triples": [
            ["http://onto-ns.com/meta/1.0/Foo#a", "map:mapsTo", "onto:A"],
            ["http://onto-ns.com/meta/1.0/Foo#b", "map:mapsTo", "onto:B"],
            ["http://onto-ns.com/meta/1.0/Bar#a", "map:mapsTo", "onto:C"],
        ],
    },
    "transformation": {"celery_task_id": "some_task_id"},
}


class ResourceType(str, Enum):
    """Enumeration of resource types."""

    DATARESOURCE = "dataresource"
    FILTER = "filter"
    FUNCTION = "function"
    MAPPING = "mapping"
    SESSION = "session"
    TRANSFORMATION = "transformation"

    def get_idprefix(self) -> str:
        """Get the `IDPREFIX` used in oteapi-services."""
        return f"{self.value}-"

    def get_class_name(self) -> str:
        """Get the internal Python class name for the resource type."""
        if self == self.SESSION:
            raise ValueError("SESSION does not have a class name.")
        return {"dataresource": "DataResource"}.get(
            self.value, str(self.value).capitalize()
        )

    def get_return_id_key(self) -> str:
        """Return the expected key used to store in OTE Services session cache."""
        if self == self.DATARESOURCE:
            return "resource_id"
        return f"{self.value}_id"

    def map_method_to_data(self, method: "Literal['get', 'initialize']") -> bool:
        """Return whether a given method should return non-empty (test) data."""
        try:
            return (
                self
                in {
                    "get": [self.DATARESOURCE, self.FUNCTION, self.TRANSFORMATION],
                    "initialize": [self.FILTER, self.MAPPING],
                }[method]
            )
        except KeyError as exc:
            raise ValueError(
                "Only 'get' and 'initialize' methods are allowed."
            ) from exc


def strategy_create_kwargs() -> "List[Tuple[str, Dict[str, Any]]]":
    """Strategy to creation key-word-arguments."""
    return [
        (
            ResourceType.DATARESOURCE.value,
            {
                "downloadUrl": "https://filesamples.com/samples/code/json/sample2.json",
                "mediaType": "application/json",
            },
        ),
        (
            ResourceType.FILTER.value,
            {
                "filterType": "filter/sql",
                "query": TEST_DATA[ResourceType.FILTER.value]["sqlquery"],
            },
        ),
        (
            ResourceType.FUNCTION.value,
            {
                "functionType": "function/demo",
                **TEST_DATA[ResourceType.FUNCTION.value],
            },
        ),
        (
            ResourceType.MAPPING.value,
            {
                "mappingType": "triples",
                **TEST_DATA[ResourceType.MAPPING.value],
            },
        ),
        (
            ResourceType.TRANSFORMATION.value,
            {
                "transformationType": "celery/remote",
                "configuration": {
                    "task_name": "test-task",
                    "args": [],
                },
            },
        ),
    ]



"""Base API for backend strategies."""
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from otelib.backends.utils import StrategyType
from otelib.pipe import Pipe

if TYPE_CHECKING:  # pragma: no cover
    from typing import Optional, Type, Union

    from oteapi.models.genericconfig import GenericConfig


class AbstractBaseStrategy(ABC):
    """The abstract base class defining the API for strategies."""

    strategy_name: "Union[StrategyType, str]"
    strategy_config: "Type[GenericConfig]"

    def __init__(self, source: str) -> None:
        """Initiates a strategy."""
        if not source:
            raise ValueError("source must be provided.")

        self.input_pipe: "Optional[Pipe]" = None
        self.strategy_id: str = ""
        self.strategy_name = StrategyType(self.strategy_name)

        # For debugging/testing
        self.debug = bool(os.getenv("OTELIB_DEBUG", ""))
        self._session_id: "Optional[str]" = None

    @abstractmethod
    def create(self, **kwargs) -> None:
        """Create a strategy.

        It should post the configuration for the created strategy.
        """

    @abstractmethod
    def fetch(self, session_id: str) -> bytes:
        """Returns the result of the current strategy.

        This method is called by `get()` while propagating up the pipeline.

        Parameters:
            session_id: The ID of the session shared by the pipeline.

        Returns:
            The result of calling the `get()` method on the current strategy

        """

    @abstractmethod
    def initialize(self, session_id: str) -> bytes:
        """Initialise the current strategy.

        This method is called by `get()` when propagating down the pipeline.

        Parameters:
            session_id: The ID of the session shared by the pipeline.

        Returns:
            The response from the OTEAPI Service.

        """

    def get(self, session_id: "Optional[str]" = None) -> bytes:
        """Executes a pipeline.

        This will call `initialize()` and then the `get()` method on the
        input pipe, which in turn will call the `get()` method on the
        strategy connected to its input and so forth until the beginning
        of the pipeline.

        Finally, `fetch()` is called and its output is returned.

        Parameters:
            session_id: The ID of the session shared by the pipeline.

        Returns:
            The output from `fetch()`.

        """
        if session_id is None:
            session_id = self._create_session()

        if self.debug:
            self._session_id = session_id

        self.initialize(session_id)
        if self.input_pipe:
            self.input_pipe.get(session_id)
        return self.fetch(session_id)

    @abstractmethod
    def _create_session(self) -> str:
        """Create a new session.

        This method should not be run by a user, hence it is "private".
        The method is used within the `get()` method and allows a backend to customize
        its session creation method.

        Returns:
            The newly created session's ID.

        """

    def _set_input(self, input_pipe: Pipe) -> None:
        """Used by `__rshift__` to set the input pipe.

        Parameters:
            input_pipe: A pipe representing the strategy that is "piped" into this
                strategy.

        """
        start_filter = _find_start_filter(self)
        start_filter.input_pipe = input_pipe

    def __rshift__(self, other: "AbstractBaseStrategy") -> "AbstractBaseStrategy":
        """Implements strategy concatenation using the `>>` symbol.

        Parameters:
            other: The next strategy this one is "piping" into.

        Returns:
            The next strategy this one is "piped" into.

        """
        pipe = Pipe(self)
        other._set_input(pipe)
        return other


def _find_start_filter(other):
    """Used by _set_input to find the input filter,
    incase of pipeline concatenation."""

    while hasattr(other, "input_pipe"):
        pipe = other.input_pipe
        if hasattr(pipe, "input"):
            other = pipe.input
        else:
            return other
    return other



"""OTElib exceptions."""


class BaseOtelibException(Exception):
    """A base OTElib exception."""


class ApiError(BaseOtelibException):
    """An API Error Exception"""

    def __init__(self, detail: str, status: int, *args) -> None:
        super().__init__(detail, *args)
        self.detail = detail
        self.status = status

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: status={self.status} {self.detail}"


class InvalidBackend(BaseOtelibException):
    """The backend does not exist; it is invalid."""


class InvalidStrategy(BaseOtelibException):
    """The strategy type does not exist; it is invalid."""


class PythonBackendException(BaseOtelibException):
    """A generic error has happened in the Python backend."""


class PythonCacheError(PythonBackendException):
    """An error occurred when dealing with the cache in the Python backend."""


class ItemNotFoundInCache(PythonCacheError):
    """An item could not be found in the cache."""

    def __init__(self, detail: str, item: str, *args) -> None:
        super().__init__(detail, *args)
        self.detail = detail
        self.item = item

    def __str__(self) -> str:
        return f" {self.__class__.__name__}: item={self.item!r} {self.detail}"



"""Test OTE Client."""
# pylint: disable=protected-access,invalid-name,too-many-arguments,too-many-locals
# pylint: disable=too-many-branches
from typing import TYPE_CHECKING

import pytest
from utils import strategy_create_kwargs

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Union

    from otelib.backends.python.base import BasePythonStrategy
    from otelib.backends.services.base import BaseServicesStrategy
    from otelib.client import OTEClient

    from .conftest import OTEResponse, ResourceType

    BaseStrategy = Union[BasePythonStrategy, BaseServicesStrategy]


@pytest.mark.parametrize(
    "strategy,create_kwargs",
    strategy_create_kwargs(),
    ids=[_[0] for _ in strategy_create_kwargs()],
)
@pytest.mark.usefixtures("mock_session")
def test_create_strategies(
    client: "OTEClient",
    ids: "Callable[[Union[ResourceType, str]], str]",
    mock_ote_response: "OTEResponse",
    testdata: "Callable[[Union[ResourceType, str]], dict]",
    strategy: str,
    create_kwargs: "Dict[str, Any]",
) -> None:
    """Test creating any strategy and calling it's `get()` method."""
    import json

    import requests

    if strategy == "function":
        if client._impl._backend == "services" and "example" in client.url:
            pass
        else:
            pytest.skip("No function strategy exists in oteapi-core yet.")

    backend = client._impl._backend

    if backend == "services":
        # Mock URL responses

        # create()
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy}",
            return_json={
                f"{strategy[len('data'):] if strategy.startswith('data') else strategy}"
                "_id": ids(strategy)
            },
        )

        # initialize()
        # The filter and mapping returns everything from their `initialize()` method.
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy}/{ids(strategy)}/initialize",
            params={"session_id": ids("session")},
            return_json=(
                testdata(strategy) if strategy in ("filter", "mapping") else {}
            ),
        )

        # fetch()
        # The data resource and transformation returns everything from their `get()`
        # method.
        mock_ote_response(
            method="get",
            endpoint=f"/{strategy}/{ids(strategy)}",
            params={"session_id": ids("session")},
            return_json=(
                testdata(strategy)
                if strategy in ("dataresource", "transformation")
                else {}
            ),
        )

        # Session content
        mock_ote_response(
            method="get",
            endpoint=f"/session/{ids('session')}",
            return_json=testdata(strategy),
        )

    created_strategy: "BaseStrategy" = getattr(client, f"create_{strategy}")(
        **create_kwargs
    )

    content = created_strategy.get()
    if strategy in ("filter", "mapping"):
        assert json.loads(content) == {}
    elif (
        strategy in ("transformation",)
        and client._impl._backend == "services"
        and "example" not in client.url
    ):
        # Real backend !
        # Dynamic response content - just check keys are the same and values are
        # non-empty
        _content: "dict[str, Any]" = json.loads(content)
        assert list(_content) == list(testdata(strategy))
        assert all(_content.values())
    else:
        assert json.loads(content) == testdata(strategy)

    # The testdata should always be in the full session
    assert (
        created_strategy._session_id
    ), f"Session ID not found in {created_strategy} ! Is OTEAPI_DEBUG not set?"

    if backend == "services":
        strategy_prefix = created_strategy.settings.prefix
        startegy_sessionid = created_strategy._session_id
        content_session = requests.get(
            f"{created_strategy.url}{strategy_prefix}/session/{startegy_sessionid}",
            timeout=30,
        )
        session: "Dict[str, Any]" = content_session.json()

    elif backend == "python":
        session_ids = [
            cache_key for cache_key in created_strategy.cache if "session" in cache_key
        ]
        assert len(session_ids) == 1
        session = created_strategy.cache[session_ids[0]]

    for key, value in testdata(strategy).items():
        assert key in session

        if strategy == "mapping" and key == "triples":
            # The mapping strategy's "triples" key has a Set type value
            session_triples = sorted(list(triple) for triple in session[key])
            assert sorted(value) == session_triples
        elif (
            strategy == "transformation"
            and key == "celery_task_id"
            and "example" not in client.url
        ):
            # The task ID is dynamically generated.
            # Simply check the value is non-empty
            assert key in session and session[key]
        else:
            assert value == session[key]



"""Fixtures and configuration for pytest."""
# pylint: disable=too-many-arguments,protected-access
from enum import Enum
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Literal, Optional, Type, Union

    from requests_mock import Mocker
    from utils import ResourceType

    from otelib.client import OTEClient

    OTEResponse = Callable[
        [
            Union["HTTPMethod", str],
            str,
            int,
            Optional[Union[Dict[str, Any], str]],
            Optional[dict],
            Optional[bytes],
            Optional[Union[dict, str]],
            Optional[str],
            Optional[OTEClient],
        ],
        None,
    ]


class HTTPMethod(str, Enum):
    """Allowed HTTP methods.

    See `requests_mock.mocker` for the reference list.
    """

    DELETE = "delete"
    GET = "get"
    HEAD = "head"
    OPTIONS = "options"
    PATCH = "patch"
    POST = "post"
    PUT = "put"


def pytest_configure(config) -> None:  # pylint: disable=unused-argument
    """Method that runs before pytest collects tests, so no modules are imported."""
    import os

    os.environ["OTELIB_DEBUG"] = "True"


@pytest.fixture
def server_url() -> str:
    """Return a possibly set real server URL.

    The server URL must be set through the environment variable
    `OTELIB_TEST_OTE_SERVER_URL`.
    """
    import os

    return os.getenv("OTELIB_TEST_OTE_SERVER_URL", "https://example.org")


@pytest.fixture
def resource_type_cls() -> "Type[ResourceType]":
    """Return the `ResourceType` Enum."""
    from utils import ResourceType

    return ResourceType


@pytest.fixture
def ids() -> "Callable[[Union[ResourceType, str]], str]":
    """Provide a function to return a test resource id.

    By "resource", any resource is meant, e.g., `sessions`, `filter`, etc.
    """
    from utils import ResourceType

    def _ids(resource_type: "Union[ResourceType, str]") -> str:
        """Return a test id for the given `resource_type`."""
        resource_type = ResourceType(resource_type)
        return f"{resource_type.get_idprefix()}test"

    return _ids


@pytest.fixture(params=["services", "python"])
def backend(request: pytest.FixtureRequest) -> str:
    """Run a test for all backends."""
    if request.param == "python":
        from oteapi.plugins import load_strategies

        load_strategies()

    return request.param


@pytest.fixture
def client(server_url: str, backend: str) -> "OTEClient":
    """Create an `OTEClient` test client."""
    from otelib.client import OTEClient

    if backend == "services":
        return OTEClient(server_url)

    if backend == "python":
        res = OTEClient("python")
        res._impl.clear_cache()  # pylint: disable=no-member
        return res

    raise RuntimeError(f"Unknown backend: {backend!r}")


@pytest.fixture
def mock_session(
    requests_mock: "Mocker",
    client: "OTEClient",
    ids: "Callable[[Union[ResourceType, str]], str]",
    server_url: str,
) -> None:
    """Mock `POST /session`.

    This is called in `AbstractStrategy.get()`.
    """
    from otelib.settings import Settings

    if client._impl._backend == "services" and "example" in server_url:
        requests_mock.post(
            f"{client.url}{Settings().prefix}/session",
            json={"session_id": ids("session")},
        )
    else:
        # Make sure the requests are done for real.
        requests_mock.real_http = True


@pytest.fixture
def mock_ote_response(
    requests_mock: "Mocker", server_url: str
) -> "Callable[[Union[HTTPMethod, str], str, int, Optional[Union[Dict[str, Any], str]], Optional[dict], Optional[bytes], Optional[Union[dict, str]], Optional[str], Optional[OTEClient]], None]":  # pylint: disable=line-too-long
    """Provide a function to mock OTE services responses."""
    from urllib.parse import parse_qs

    from otelib.settings import Settings

    def _mock_response(
        method: "Union[HTTPMethod, str]",
        endpoint: str,
        status_code: int = 200,
        params: "Optional[Union[Dict[str, Any], str]]" = None,
        headers: "Optional[dict]" = None,
        return_content: "Optional[bytes]" = None,
        return_json: "Optional[Union[dict, str]]" = None,
        return_text: "Optional[str]" = None,
        ote_client: "Optional[OTEClient]" = None,
    ) -> None:
        """Use `requests_mock` to mock a response from an OTE services server.

        It will only be ensured that the `endpoint` starts with a forward slash.
        If it does not, one will be added. Otherwise, `endpoint` is not manipulated.
        """
        if "example" not in server_url:
            # Make sure the requests are done for real.
            requests_mock.real_http = True
            return

        method: str = (
            HTTPMethod(method.lower())
            if isinstance(method, str)
            else HTTPMethod(method)
        ).value.upper()

        if not endpoint.startswith("/"):
            endpoint = f"/{endpoint}"

        if params:
            if isinstance(params, str):
                params = parse_qs(params)
            _params = "?"

            key = sorted(params)[0]
            values = (
                [params.pop(key)]
                if isinstance(params[key], (str, int, float))
                else params.pop(key)
            )
            _params += "&".join(f"{key}={value}" for value in sorted(values))
            while params:
                key = sorted(params)[0]
                values = (
                    [params.pop(key)]
                    if isinstance(params[key], (str, int, float))
                    else params.pop(key)
                )
                _params += "&" + "&".join(f"{key}={value}" for value in sorted(values))

            params = _params
        else:
            params = ""

        mock_kwargs = {}
        if return_content is not None:
            mock_kwargs["content"] = return_content
        if return_json is not None:
            mock_kwargs["json"] = return_json
        if return_text is not None:
            mock_kwargs["text"] = return_text
        if headers is not None:
            mock_kwargs["headers"] = headers

        print(
            f"Mocking request: {method} "
            f"{ote_client.url if ote_client else server_url}{Settings().prefix}"
            f"{endpoint}{params} "
            f"status_code={status_code} "
            + " ".join(f"{key}={value}" for key, value in mock_kwargs.items())
        )

        requests_mock.request(
            method=method,
            url=(
                f"{ote_client.url if ote_client else server_url}{Settings().prefix}"
                f"{endpoint}{params}"
            ),
            status_code=status_code,
            **mock_kwargs,
        )

    return _mock_response


@pytest.fixture
def raw_test_data() -> "Dict[str, Any]":
    """Return raw test data."""
    from copy import deepcopy

    from utils import TEST_DATA

    return deepcopy(TEST_DATA)


@pytest.fixture
def testdata(
    raw_test_data: "Dict[str, Any]",
) -> "Callable[[Union[ResourceType, str]], dict]":
    """Test data for OTE resource."""
    from utils import ResourceType

    def _testdata(
        resource_type: "Union[ResourceType, str]",
        method: "Optional[Literal['get', 'initialize']]" = None,
    ) -> dict:
        """Return test data for a given resource.

        Parameters:
            resource_type: The resource type to return test data for.
            method: Optionally, provide the method for which to return the test data.
                If not given, the test data will be provided as is.

        """
        resource_type = ResourceType(resource_type)
        if resource_type == ResourceType.SESSION:
            raise ValueError("No test data available for a session.")

        if method is None:
            return raw_test_data[resource_type.value]

        return (
            raw_test_data[resource_type.value]
            if resource_type.map_method_to_data(method)
            else {}
        )

    return _testdata


@pytest.fixture(autouse=True)
def mock_celery_transformation_strategy(
    monkeypatch: pytest.MonkeyPatch, raw_test_data: "Dict[str, Any]"
) -> None:
    """Use celery_worker always for all things.

    Parameters:
        monkeypatch: Monkeypatch fixture for use in pytest.
        raw_test_data: Deep copy of the test data.

    """

    class MockResult:
        """Mock result from 'transformationType=celery/remote'."""

        task_id = raw_test_data["transformation"]["celery_task_id"]

    monkeypatch.setattr(
        "oteapi.strategies.transformation.celery_remote.CELERY_APP.send_task",
        lambda *args, **kwargs: MockResult(),
    )



"""Pytest fixtures for everything in tests/strategies."""
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from typing import Literal, Tuple, Type, Union

    from otelib.backends import python as python_backend
    from otelib.backends import services as services_backend

    from ..conftest import ResourceType

    StrategyCls = Union[
        python_backend.DataResource,
        python_backend.Filter,
        python_backend.Function,
        python_backend.Mapping,
        python_backend.Transformation,
        services_backend.DataResource,
        services_backend.Filter,
        services_backend.Function,
        services_backend.Mapping,
        services_backend.Transformation,
    ]


@pytest.fixture(
    params=["dataresource", "filter", "function", "mapping", "transformation"]
)
def strategy_implementation(
    request: pytest.FixtureRequest,
    backend: str,
    resource_type_cls: "Type[ResourceType]",
) -> "Tuple[Type[StrategyCls], ResourceType, Literal['python', 'services']]":
    """Return a given strategy class implementation as well as the current backend.

    Returns:
        Tuple of the backend-specific strategy class, the `ResourceType` Enum
        representation of the strategy, and the given backend.

    """
    import importlib

    strategy = resource_type_cls(request.param)

    module = importlib.import_module(f"otelib.backends.{backend}.{strategy.value}")
    return getattr(module, strategy.get_class_name()), strategy, backend



"""OTE Client."""
from typing import TYPE_CHECKING

from pydantic import AnyHttpUrl, ValidationError, parse_obj_as

from otelib.backends.factories import client_factory
from otelib.backends.utils import Backend, StrategyType

if TYPE_CHECKING:  # pragma: no cover
    from otelib.backends.strategies import AbstractBaseStrategy


class OTEClient:
    """The OTEClient object representing a remote OTE REST API.

    Parameters:
        url (str): The base URL of the OTEAPI Service.

    Attributes:
        url (str): The base URL of the OTEAPI Service.

    """

    def __init__(self, url: str, **config) -> None:
        """Initialize an OTE Client.

        Parameters:
            url: The base URL of the OTE Service (or Python interpreter for local
                OTEAPI Core usage).
            config: Custom client configuration properties.

        """
        try:
            parse_obj_as(AnyHttpUrl, url)
        except ValidationError:
            backend = Backend.PYTHON
        else:
            backend = Backend.SERVICES

        self._impl = client_factory(backend)(url, **config)

    @property
    def url(self) -> str:
        """The base URL of the OTE Service.

        This may also be the Python interpreter for local OTEAPI Core usage.
        """
        return self._impl.source

    def create_dataresource(self, **config) -> "AbstractBaseStrategy":
        """Create a new data resource.

        Any given keyword arguments are passed on to the `create_strategy()` method.

        Returns:
            The newly created data resource.

        """
        return self._impl.create_strategy(StrategyType.DATARESOURCE, **config)

    def create_filter(self, **config) -> "AbstractBaseStrategy":
        """Create a new filter.

        Any given keyword arguments are passed on to the `create_strategy()` method.

        Returns:
            The newly created filter.

        """
        return self._impl.create_strategy(StrategyType.FILTER, **config)

    def create_function(self, **config) -> "AbstractBaseStrategy":
        """Create a new function.

        Any given keyword arguments are passed on to the `create_strategy()` method.

        Returns:
            The newly created function.

        """
        return self._impl.create_strategy(StrategyType.FUNCTION, **config)

    def create_mapping(self, **config) -> "AbstractBaseStrategy":
        """Create a new mapping.

        Any given keyword arguments are passed on to the `create_strategy()` method.

        Returns:
            The newly created mapping.

        """
        return self._impl.create_strategy(StrategyType.MAPPING, **config)

    def create_transformation(self, **config) -> "AbstractBaseStrategy":
        """Create a new transformation.

        Any given keyword arguments are passed on to the `create_strategy()` method.

        Returns:
            The newly created transformation.

        """
        return self._impl.create_strategy(StrategyType.TRANSFORMATION, **config)



"""Utility function and classes for use in the Backends module."""
from enum import Enum


class Backend(str, Enum):
    """Backend enumeration."""

    PYTHON = "python"
    SERVICES = "services"


class StrategyType(str, Enum):
    """Enumeration of strategy types."""

    DATARESOURCE = "dataresource"
    FILTER = "filter"
    FUNCTION = "function"
    MAPPING = "mapping"
    TRANSFORMATION = "transformation"

    @property
    def cls_name(self) -> str:
        """Return Python class name."""
        return {self.DATARESOURCE: "DataResource"}.get(self, self.capitalize())



"""OTElib warnings."""


class BaseOtelibWarning(Warning):
    """A base OTElib warning."""


class IgnoringConfigOptions(BaseOtelibWarning):
    """Some given configuration option(s) for the client is/are ignored."""



"""Client for python backend."""
from typing import TYPE_CHECKING

from oteapi.plugins import load_strategies

from otelib.backends.client import AbstractBaseClient
from otelib.exceptions import PythonBackendException

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Dict, Type

    from otelib.backends.python.base import BasePythonStrategy

CACHE: "Dict[str, Any]" = {}


# pylint: disable=duplicate-code
class OTEPythonClient(AbstractBaseClient):
    """The Python version of the OTEClient object.

    Attributes:
        interpreter (str): Interpreter for the python backend.

    """

    _backend = "python"

    def __init__(self, source: str, **config) -> None:
        """Initiates an OTEAPI Python client."""
        super().__init__(source, **config)

        self._cache = CACHE

        load_strategies()

    @property
    def interpreter(self) -> str:
        """Proxy for the source attribute."""
        return self.source

    def _validate_source(self, source: str) -> None:
        if source != "python":
            raise NotImplementedError(
                f"{source!r} is not a valid Python backend interpreter source. "
                "Only the 'python' interpreter is supported for the Python backend."
            )
        super()._validate_source(source)

    def _create_strategy(  # type: ignore[override]
        self, strategy_cls: "Type[BasePythonStrategy]", **config
    ) -> "BasePythonStrategy":
        strategy = strategy_cls(self.interpreter, self._cache)
        strategy.create(**config)
        return strategy

    def clear_cache(self) -> None:
        """Clear the global CACHE object."""
        global CACHE  # pylint: disable=global-statement
        CACHE = {}
        if self._cache != CACHE and (self._cache or CACHE):
            raise PythonBackendException("Could not clear the global CACHE object.")



"""Pytest for testing kwargs in clients"""
import pytest


def test_warning(backend: str, server_url: str) -> None:
    """Pytest for unloaded kwargs warning."""
    import warnings

    from otelib import OTEClient
    from otelib.warnings import IgnoringConfigOptions

    valid_config = {"headers": {"foo": "bar"}}
    invalid_config = valid_config["headers"]
    source = backend if backend == "python" else server_url

    if backend == "services":
        # no warning should be emitted
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            OTEClient(source, **valid_config)
    elif backend == "python":
        # headers are not supported for the Python backend
        with pytest.warns(IgnoringConfigOptions, match=r"^The given configuration.*"):
            OTEClient(source, **valid_config)
    else:
        pytest.fail(f"Unknown backend: {backend}")

    # invalid configuration
    with pytest.warns(IgnoringConfigOptions, match=r"^The given configuration.*"):
        OTEClient(source, **invalid_config)



"""Tests for `otelib.strategies.abc`."""
# pylint: disable=too-many-arguments,too-many-locals,protected-access
from typing import TYPE_CHECKING

import pytest
from utils import strategy_create_kwargs

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Union

    from requests_mock import Mocker

    from otelib.backends.python.base import BasePythonStrategy
    from otelib.backends.services.base import BaseServicesStrategy

    from ..conftest import OTEResponse, ResourceType

    BaseStrategy = Union[BasePythonStrategy, BaseServicesStrategy]


@pytest.mark.parametrize(
    "strategy_name,create_kwargs",
    strategy_create_kwargs(),
    ids=[_[0] for _ in strategy_create_kwargs()],
)
@pytest.mark.usefixtures("mock_session")
def test_get(
    backend: str,
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    testdata: "Callable[[Union[ResourceType, str]], dict]",
    server_url: str,
    strategy_name: str,
    create_kwargs: "Dict[str, Any]",
) -> None:
    """Test `AbstractStrategy.get()`."""
    if strategy_name == "function":
        pytest.skip("No function strategy exists in oteapi-core yet.")

    import importlib
    import json

    import requests

    strategies = importlib.import_module(f"otelib.backends.{backend}")
    server_url = server_url if backend != "python" else backend

    if backend == "services":
        # Mock URL responses

        ## create()
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_name}",
            return_json={
                f"{strategy_name[len('data'):] if strategy_name.startswith('data') else strategy_name}"  # pylint: disable=line-too-long
                "_id": ids(strategy_name)
            },
        )

        # initialize()
        # The filter and mapping returns everything from their `initialize()` method.
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_name}/{ids(strategy_name)}/initialize",
            params={"session_id": ids("session")},
            return_json=(
                testdata(strategy_name)
                if strategy_name in ("filter", "mapping")
                else {}
            ),
        )

        # fetch()
        # The data resource and transformation returns everything from their `get()`
        # method.
        mock_ote_response(
            method="get",
            endpoint=f"/{strategy_name}/{ids(strategy_name)}",
            params={"session_id": ids("session")},
            return_json=(
                testdata(strategy_name)
                if strategy_name in ("dataresource", "transformation")
                else {}
            ),
        )

        # Session content
        mock_ote_response(
            method="get",
            endpoint=f"/session/{ids('session')}",
            return_json=testdata(strategy_name),
        )

    strategy_name_map = {"dataresource": "DataResource"}

    strategy: "BaseStrategy" = getattr(
        strategies, strategy_name_map.get(strategy_name, strategy_name.capitalize())
    )(server_url)

    # We must create the strategy - getting a strategy ID
    strategy.create(**create_kwargs)
    assert strategy.strategy_id

    # There must be a strategy name associated with the strategy
    assert strategy.strategy_name == strategy_name

    content = strategy.get()
    assert isinstance(content, bytes)
    if strategy_name in ("filter", "mapping"):
        assert json.loads(content) == {}
    elif (
        strategy_name in ("transformation",)
        and backend == "services"
        and "example" not in server_url
    ):
        # Real backend !
        # Dynamic response content - just check keys are the same and values are
        # non-empty
        _content: "dict[str, Any]" = json.loads(content)
        assert list(_content) == list(testdata(strategy_name))
        assert all(_content.values())
    else:
        assert json.loads(content) == testdata(strategy_name)

    ## The testdata should always be in the full session
    assert (
        strategy._session_id
    ), f"Session ID not found in {strategy_name} ! Is OTEAPI_DEBUG not set?"

    if backend == "services":
        content_session = requests.get(
            f"{strategy.url}{strategy.settings.prefix}/session/{strategy._session_id}",
            timeout=30,
        )
        session: "Dict[str, Any]" = content_session.json()
    elif backend == "python":
        session_ids = [x for x in strategy.cache if "session" in x]
        assert len(session_ids) == 1
        session_id = session_ids[0]
        session = strategy.cache[session_id]

    for key, value in testdata(strategy_name).items():
        assert key in session

        if strategy_name == "mapping" and key == "triples":
            # The mapping strategy's "triples" key has a Set type value
            session_triples = sorted(list(triple) for triple in session[key])
            assert sorted(value) == session_triples
        elif (
            strategy_name == "transformation"
            and key == "celery_task_id"
            and "example" not in server_url
        ):
            # The task ID is dynamically generated.
            # Simply check the value is non-empty
            assert key in session and session[key]
        else:
            assert value == session[key]


@pytest.mark.parametrize(
    "strategy_name,create_kwargs",
    strategy_create_kwargs(),
    ids=[_[0] for _ in strategy_create_kwargs()],
)
def test_services_get_fails(
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    server_url: str,
    strategy_name: str,
    create_kwargs: "Dict[str, Any]",
    requests_mock: "Mocker",
) -> None:
    """Check `AbstractStrategy.get()` raises `ApiError` upon request failure."""
    from otelib.backends import services as strategies
    from otelib.exceptions import ApiError
    from otelib.settings import Settings

    wrong_url = "https://non.existing.url.org"

    # create()
    mock_ote_response(
        method="post",
        endpoint=f"/{strategy_name}",
        return_json={
            f"{strategy_name[len('data'):] if strategy_name.startswith('data') else strategy_name}"  # pylint: disable=line-too-long
            "_id": ids(strategy_name)
        },
    )

    # Creating a session
    requests_mock.post(
        f"{wrong_url}{Settings().prefix}/session",
        status_code=500,
        content=b"Internal Server Error",
    )

    strategy_name_map = {"dataresource": "DataResource"}

    strategy: "BaseStrategy" = getattr(
        strategies, strategy_name_map.get(strategy_name, strategy_name.capitalize())
    )(server_url)

    # We must create the strategy - getting a strategy ID
    strategy.create(**create_kwargs)
    assert strategy.strategy_id

    with pytest.raises(ApiError, match=f"^{ApiError.__name__}.*"):
        # Change `url` attribute to hit a wrong URL and raise
        strategy.url = wrong_url
        strategy.get()



"""Base class for strategies in the Python backend."""
import json
import warnings
from copy import deepcopy
from typing import TYPE_CHECKING
from uuid import uuid4

from oteapi.plugins import create_strategy

from otelib.backends.strategies import AbstractBaseStrategy
from otelib.exceptions import ItemNotFoundInCache, PythonBackendException

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Dict, Literal, Optional

    from oteapi.models import SessionUpdate


class BasePythonStrategy(AbstractBaseStrategy):
    """Base class for strategies for the python backend.

    Parameters:
        source (str): The Python interpreter to use in the local environment.
            Currently only `python` is allowed.

    Attributes:
        interpreter (str): This is always `python` for the Python backend.
        input_pipe (Optional[Pipe]): An input pipeline.

    """

    def __init__(self, source: str, cache: "Optional[Dict[str, Any]]" = None) -> None:
        super().__init__(source)

        self.interpreter: "Optional[str]" = source
        if cache is None:
            warnings.warn(f"No global cache used for Python backend strategy {self}")
        self.cache = cache if cache is not None else {}

        if self.interpreter != "python":
            raise ValueError(
                "Only the 'python' interpreter source is currently supported."
            )

    def create(self, **config) -> None:
        session_id = config.pop("session_id", None)
        data = self.strategy_config(**config)

        self.strategy_id = f"{self.strategy_name}-{uuid4()}"
        self.cache[self.strategy_id] = data.json()

        if session_id:
            if session_id not in self.cache:
                raise ItemNotFoundInCache(
                    "Session not found in cache, but given to create()", session_id
                )

            # Add strategy ID information to the session object.
            list_key = f"{self.strategy_name}_info"
            if list_key in self.cache[session_id]:
                if not isinstance(self.cache[session_id][list_key], list):
                    raise TypeError(
                        f"Expected type for {list_key!r} field in session to be a "
                        f"list, found {type(self.cache[session_id][list_key])!r}."
                    )
                self.cache[session_id][list_key].append(self.strategy_id)
            else:
                self.cache[session_id][list_key] = [self.strategy_id]

    def fetch(self, session_id: str) -> bytes:
        return self._run_strategy_method("get", session_id)

    def initialize(self, session_id: str) -> bytes:
        return self._run_strategy_method("initialize", session_id)

    def _create_session(self) -> str:
        session_id = f"session-{uuid4()}"
        self.cache[session_id] = {}
        return session_id

    def _run_strategy_method(
        self, method_name: "Literal['get', 'initialize']", session_id: str
    ) -> bytes:
        """Generic implementation of the `fetch()` and `initialize()` methods.

        This will run the `method_name` method on the strategy and return the
        serialized session update object.

        Parameters:
            method_name: The name of the strategy's method to execute.
            session_id: The ID of the session shared by the pipeline.

        Returns:
            The bytes-serialized output from the given strategy method.

        """
        if method_name not in ["get", "initialize"]:
            raise PythonBackendException(
                "method_name should be either 'get' or 'initialize'."
            )

        self._sanity_checks(session_id)

        config = self.strategy_config(**json.loads(self.cache[self.strategy_id]))
        strategy = create_strategy(self.strategy_name, config)

        if not hasattr(strategy, method_name):
            raise PythonBackendException(
                f"{method_name!r} is not a valid method for {strategy}"
            )

        session_update: "SessionUpdate" = getattr(strategy, method_name)(
            session=deepcopy(self.cache[session_id])
        )

        self.cache[session_id].update(session_update)

        return session_update.json().encode(encoding="utf-8")

    def _sanity_checks(self, session_id: str) -> None:
        """Perform sanity checks before running a strategy method.

        Parameters:
            session_id: The ID of the session shared by the pipeline.

        """
        if not self.strategy_id or self.strategy_id not in self.cache:
            raise ItemNotFoundInCache(
                "Run create() prior to initialize()", self.strategy_id
            )

        if session_id not in self.cache or not isinstance(
            self.cache.get(session_id, {}), dict
        ):
            raise ItemNotFoundInCache(
                "Did you run this method through get()?", session_id
            )



"""Base API for backend client."""
import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from otelib.backends.factories import strategy_factory
from otelib.backends.utils import Backend, StrategyType
from otelib.warnings import IgnoringConfigOptions

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Dict, Type, Union

    from otelib.backends.strategies import AbstractBaseStrategy


class AbstractBaseClient(ABC):
    """The abstract base class defining the API for a backend client."""

    _backend: "Union[Backend, str]"

    def __init__(self, source: str, **config) -> None:
        """Initiates a client."""
        if not source:
            raise ValueError("source must be provided.")

        self._source = ""
        self._backend = Backend(self._backend)

        self._validate_source(source)
        self._set_config(config)

    @property
    def source(self) -> str:
        """The backend/source the client is using.

        This may be an OTE Services base URL or a Python interpreter.
        """
        return self._source

    def _validate_source(self, source: str) -> None:
        """Validate the `source` initialization variable.

        This method should be overridden in the given implementation, and finally,
        _this_ parent method must be called to set the `source` attribute.
        """
        if self.source:
            raise RuntimeError(
                "'source' should not be set prior to calling '_validate_source()'."
            )
        self._source = source

    def _set_config(self, config: "Dict[str, Any]") -> None:
        """Set the custom client configuration options."""
        if config:
            warnings.warn(
                f"The given configuration option(s) for {tuple(config)} is/are "
                "ignored.",
                IgnoringConfigOptions,
            )

    @abstractmethod
    def _create_strategy(
        self, strategy_cls: "Type[AbstractBaseStrategy]", **config
    ) -> "AbstractBaseStrategy":
        """Create a strategy.

        This method should not be run by a user, hence it is "private".
        The method is used with the `create_strategy()` method and allows a backend to
        customize its strategy creation method.

        Returns:
            The newly created strategy.

        """

    def create_strategy(
        self, strategy_type: "Union[str, StrategyType]", **config
    ) -> "AbstractBaseStrategy":
        """Create a strategy."""
        strategy_cls = strategy_factory(self._backend, strategy_type)
        return self._create_strategy(strategy_cls, **config)



"""Pipe object for creating a pipeline."""
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from typing import Optional

    from otelib.backends.strategies import AbstractBaseStrategy


class Pipe:
    """Pipe object in a pipe-and-filter pattern."""

    def __init__(self, strategy: "AbstractBaseStrategy") -> None:
        self.input: "AbstractBaseStrategy" = strategy

    def get(self, session_id: "Optional[str]" = None) -> bytes:
        """Call the input strategy's `get()` method."""
        return self.input.get(session_id)



"""Filter strategy."""
from oteapi.models import FilterConfig

from otelib.backends.services.base import BaseServicesStrategy


class Filter(BaseServicesStrategy):
    """Context class for the Filter Strategy Interfaces"""

    strategy_name = "filter"
    strategy_config = FilterConfig



"""Filter strategy."""
from oteapi.models import TransformationConfig

from otelib.backends.services.base import BaseServicesStrategy


class Transformation(BaseServicesStrategy):
    """Context class for the Transformation Strategy Interfaces"""

    strategy_name = "transformation"
    strategy_config = TransformationConfig



"""OTElib

A package to interact with the Open Translation Environment (OTE) API.
"""
from .client import OTEClient

__all__ = ("OTEClient",)

__version__ = "0.3.2"
__author__ = "SINTEF"
__author_email__ = "Team4.0@SINTEF.no"



"""The strategies to be created by the client."""
from .dataresource import DataResource
from .filter import Filter
from .function import Function
from .mapping import Mapping
from .transformation import Transformation

# pylint: disable=duplicate-code
__all__ = (
    "DataResource",
    "Filter",
    "Function",
    "Mapping",
    "Transformation",
)



"""Backend factory functions."""
import importlib
from typing import TYPE_CHECKING

from otelib.backends.utils import Backend, StrategyType
from otelib.exceptions import InvalidBackend, InvalidStrategy

if TYPE_CHECKING:  # pragma: no cover
    from typing import Type, Union

    from otelib.backends.client import AbstractBaseClient
    from otelib.backends.strategies import AbstractBaseStrategy


def client_factory(backend: "Union[str, Backend]") -> "Type[AbstractBaseClient]":
    """Return a backend client class."""
    try:
        backend = Backend(backend)
    except ValueError as exc:
        raise InvalidBackend(f"{str(backend)!r} is not a valid backend.") from exc

    client_module = importlib.import_module(f"otelib.backends.{backend}.client")

    if backend == backend.PYTHON:
        return getattr(client_module, "OTEPythonClient")

    if backend == backend.SERVICES:
        return getattr(client_module, "OTEServiceClient")

    raise InvalidBackend(f"{str(backend)!r} is not a valid backend.")


def strategy_factory(
    backend: "Union[str, Backend]", strategy_type: "Union[str, StrategyType]"
) -> "Type[AbstractBaseStrategy]":
    """Return a backend-specific strategy class."""
    try:
        backend = Backend(backend)
    except ValueError as exc:
        raise InvalidBackend(f"{str(backend)!r} is not a valid backend.") from exc

    try:
        strategy_type = StrategyType(strategy_type)
    except ValueError as exc:
        raise InvalidStrategy(
            f"{str(strategy_type)!r} is not a valid strategy."
        ) from exc

    strategy_module = importlib.import_module(
        f"otelib.backends.{backend}.{strategy_type}"
    )
    try:
        return getattr(strategy_module, strategy_type.cls_name)
    except AttributeError as exc:
        raise NotImplementedError(
            f"The {str(strategy_type)!r} is (currently) not supported by the "
            f"{str(backend)!r} backend."
        ) from exc



"""Generic tests for all strategies."""
# pylint: disable=too-many-arguments,protected-access
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from typing import Any, Callable, Tuple, Type, Union

    from requests_mock import Mocker
    from utils import ResourceType

    from ..conftest import OTEResponse
    from .conftest import StrategyCls


def test_create(
    strategy_implementation: "Tuple[Type[StrategyCls], ResourceType, str]",
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    server_url: str,
) -> None:
    """Test the `create()` method."""
    from utils import strategy_create_kwargs

    strategy_cls, strategy_type, backend = strategy_implementation
    server_url = server_url if backend != "python" else backend

    if backend == "services":
        # Mock URL responses
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}",
            return_json={strategy_type.get_return_id_key(): ids(strategy_type.value)},
        )

    strategy = strategy_cls(server_url)

    assert not strategy.strategy_id

    strategy.create(**dict(strategy_create_kwargs())[strategy_type.value])

    assert strategy.strategy_id


def test_create_fails(
    strategy_implementation: "Tuple[Type[StrategyCls], ResourceType, str]",
    mock_ote_response: "OTEResponse",
    server_url: str,
) -> None:
    """Check `create()` raises `ApiError` upon request failure."""
    from utils import strategy_create_kwargs

    from otelib.exceptions import ApiError

    strategy_cls, strategy_type, backend = strategy_implementation

    if backend == "python":
        pytest.skip(
            "Currently no tests for expected create() failures for "
            f"backend={backend!r}."
        )

    if backend == "services":
        # Mock URL responses
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}",
            status_code=500,
            return_content=b"Internal Server Error",
        )

    strategy = strategy_cls(server_url)

    assert not strategy.strategy_id

    valid_config = dict(strategy_create_kwargs())[strategy_type.value]
    with pytest.raises(ApiError, match=f"^{ApiError.__name__}.*"):
        # `session_id` has a wrong type, the request should fail.
        strategy.create(session_id=123, **valid_config)

    assert not strategy.strategy_id


def test_fetch(
    strategy_implementation: "Tuple[Type[StrategyCls], ResourceType, str]",
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    server_url: str,
    testdata: "Callable[[Union[ResourceType, str]], dict]",
    requests_mock: "Mocker",
) -> None:
    """Test the `fetch()` method."""
    import json

    from utils import strategy_create_kwargs

    strategy_cls, strategy_type, backend = strategy_implementation
    server_url = server_url if backend != "python" else backend

    if strategy_type == strategy_type.FUNCTION and not (
        backend == "services" and "example" in server_url
    ):
        pytest.skip("No function strategy exists in oteapi-core yet.")

    if backend == "services":
        # Mock URL responses
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}",
            return_json={strategy_type.get_return_id_key(): ids(strategy_type.value)},
        )

        mock_ote_response(
            method="get",
            endpoint=f"/{strategy_type.value}/{ids(strategy_type.value)}",
            return_json=testdata(strategy_type, "get"),
        )

    strategy = strategy_cls(server_url)

    session_id = ""
    if backend == "python":
        # Create session
        session_id = strategy._create_session()

    # We must first create the resource - getting a resource ID
    strategy.create(**dict(strategy_create_kwargs())[strategy_type.value])

    if backend == "python" and strategy_type == strategy_type.DATARESOURCE:
        # Mock URL responses
        requests_mock.request(
            method="get",
            url=dict(strategy_create_kwargs())[strategy_type.value]["downloadUrl"],
            status_code=200,
            json=testdata(strategy_type, "get")["content"],
        )

    content = strategy.fetch(session_id)

    if (
        strategy_type == strategy_type.TRANSFORMATION
        and backend == "services"
        and "example" not in server_url
    ):
        # Real backend for transformation strategy
        # Dynamic response content - just check keys are the same and values are
        # non-empty
        _content: "dict[str, Any]" = json.loads(content)
        assert list(_content) == list(testdata(strategy_type, "get"))
        assert all(_content.values())
    else:
        assert json.loads(content) == testdata(strategy_type, "get")


def test_fetch_fails(
    strategy_implementation: "Tuple[Type[StrategyCls], ResourceType, str]",
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    server_url: str,
) -> None:
    """Check `fetch()` raises `ApiError` upon request failure."""
    from utils import strategy_create_kwargs

    from otelib.exceptions import ApiError

    strategy_cls, strategy_type, backend = strategy_implementation

    if backend == "python":
        pytest.skip(
            f"Currently no tests for expected fetch() failures for backend={backend!r}."
        )

    if backend == "services":
        # Mock URL responses
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}",
            return_json={strategy_type.get_return_id_key(): ids(strategy_type.value)},
        )

        mock_ote_response(
            method="get",
            endpoint=f"/{strategy_type.value}/{ids(strategy_type.value)}",
            status_code=500,
            return_content=b"Internal Server Error",
        )

    strategy = strategy_cls(server_url)

    # We must first create the resource - getting a resource ID
    strategy.create(**dict(strategy_create_kwargs())[strategy_type.value])

    with pytest.raises(ApiError, match=f"^{ApiError.__name__}.*"):
        # `session_id` has a wrong type, the request should fail.
        strategy.fetch(session_id=123)


def test_initialize(
    strategy_implementation: "Tuple[Type[StrategyCls], ResourceType, str]",
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    server_url: str,
    testdata: "Callable[[Union[ResourceType, str]], dict]",
) -> None:
    """Test `DataResource.initialize()`."""
    import json

    from utils import strategy_create_kwargs

    strategy_cls, strategy_type, backend = strategy_implementation
    server_url = server_url if backend != "python" else backend

    if strategy_type == strategy_type.FUNCTION and not (
        backend == "services" and "example" in server_url
    ):
        pytest.skip("No function strategy exists in oteapi-core yet.")

    if backend == "services":
        # Mock URL responses
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}",
            return_json={strategy_type.get_return_id_key(): ids(strategy_type.value)},
        )

        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}/{ids(strategy_type.value)}/initialize",
            return_json=testdata(strategy_type, "initialize"),
        )

    strategy = strategy_cls(server_url)

    session_id = ""
    if backend == "python":
        # Create session
        session_id = strategy._create_session()

    # We must first create the resource - getting a resource ID
    strategy.create(**dict(strategy_create_kwargs())[strategy_type.value])

    content = strategy.initialize(session_id)

    loaded_content = json.loads(content)
    testdata_strategy = testdata(strategy_type, "initialize")
    if strategy_type == strategy_type.MAPPING and (
        backend == "python" or "example" not in strategy.url
    ):
        # "triples" is a Set and must be "sorted"
        loaded_content["triples"] = sorted(loaded_content["triples"])
        testdata_strategy["triples"] = sorted(testdata_strategy["triples"])

    assert loaded_content == testdata_strategy


def test_initialize_fails(
    strategy_implementation: "Tuple[Type[StrategyCls], ResourceType, str]",
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    server_url: str,
) -> None:
    """Check `DataResource.initialize()` raises `ApiError` upon request failure."""
    from utils import strategy_create_kwargs

    from otelib.exceptions import ApiError

    strategy_cls, strategy_type, backend = strategy_implementation

    if backend == "python":
        pytest.skip(
            "Currently no tests for expected initialize() failures for "
            f"backend={backend!r}."
        )

    if backend == "services":
        # Mock URL responses
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}",
            return_json={strategy_type.get_return_id_key(): ids(strategy_type.value)},
        )

        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_type.value}/{ids(strategy_type.value)}/initialize",
            status_code=500,
            return_content=b"Internal Server Error",
        )

    strategy = strategy_cls(server_url)

    # We must first create the resource - getting a resource ID
    strategy.create(**dict(strategy_create_kwargs())[strategy_type.value])

    with pytest.raises(ApiError, match=f"^{ApiError.__name__}.*"):
        # `session_id` has a wrong type, the request should fail.
        strategy.initialize(session_id=123)



"""Base class for strategies in the service/REST API backend."""
import json
from typing import TYPE_CHECKING

import requests

from otelib.backends.strategies import AbstractBaseStrategy
from otelib.exceptions import ApiError
from otelib.settings import Settings

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Dict, Optional


class BaseServicesStrategy(AbstractBaseStrategy):
    """Abstract class for strategies.

    Parameters:
        source (str): The base URL of the OTEAPI Service.

    Attributes:
        url (str): The base URL of the OTEAPI Service.
        settings (otelib.settings.Settings): OTEAPI Service settings.
        input_pipe (Optional[Pipe]): An input pipeline.

    """

    def __init__(self, source: str) -> None:
        super().__init__(source)

        self.url: "Optional[str]" = source
        self._headers: "Optional[Dict[str, Any]]" = None
        self.settings = Settings()

    @property
    def headers(self) -> "Dict[str, Any]":
        """URL headers to use for all requests to the OTEAPI Service."""
        value = self._headers or {}
        if "Content-Type" not in value:
            value["Content-Type"] = "application/json"
        return value

    @headers.setter
    def headers(self, value: "Dict[str, Any]") -> None:
        """Set the URL headers to use for all requests to the OTEAPI Service."""
        if not isinstance(value, dict):
            raise TypeError("headers must be a dictionary")
        self._headers = value

    def create(self, **config) -> None:
        session_id = config.pop("session_id", None)
        data = self.strategy_config(**config)

        response = requests.post(
            f"{self.url}{self.settings.prefix}/{self.strategy_name}",
            data=data.json(),
            params={"session_id": session_id} if session_id else {},
            timeout=self.settings.timeout,
            headers=self.headers,
        )
        if not response.ok:
            raise ApiError(
                f"Cannot create {self.strategy_name}: {data!r}"
                f"{' content=' + str(response.content) if self.debug else ''}",
                status=response.status_code,
            )

        response_json: dict = response.json()
        self.strategy_id = (
            response_json.pop(f"{self.strategy_name}_id")
            if f"{self.strategy_name}_id" in response_json
            else response_json.pop(f"{self.strategy_name[len('data'):]}_id")
        )

    def fetch(self, session_id: str) -> bytes:
        response = requests.get(
            f"{self.url}{self.settings.prefix}/{self.strategy_name}/{self.strategy_id}",
            params={"session_id": session_id},
            timeout=self.settings.timeout,
            headers=self.headers,
        )
        if response.ok:
            return response.content
        strategy_name = (
            self.strategy_name[len("data") :]
            if self.strategy_name.startswith("data")
            else self.strategy_name
        )
        raise ApiError(
            f"Cannot fetch {self.strategy_name}: session_id={session_id!r} "
            f"{strategy_name}_id={self.strategy_id!r}"
            f"{' content=' + str(response.content) if self.debug else ''}",
            status=response.status_code,
        )

    def initialize(self, session_id: str) -> bytes:
        post_path = (
            f"{self.url}{self.settings.prefix}"
            f"/{self.strategy_name}/{self.strategy_id}/initialize"
        )
        response = requests.post(
            post_path,
            params={"session_id": session_id},
            timeout=self.settings.timeout,
            headers=self.headers,
        )
        if response.ok:
            return response.content
        strategy_name = (
            self.strategy_name[len("data") :]
            if self.strategy_name.startswith("data")
            else self.strategy_name
        )
        raise ApiError(
            f"Cannot initialize {self.strategy_name}: session_id={session_id!r} "
            f"{strategy_name}_id={self.strategy_id!r}"
            f"{' content=' + str(response.content) if self.debug else ''}",
            status=response.status_code,
        )

    def _create_session(self) -> str:
        response = requests.post(
            f"{self.url}{self.settings.prefix}/session",
            json={},
            headers=self.headers,
            timeout=self.settings.timeout,
        )
        if not response.ok:
            raise ApiError(
                f"Cannot create session: {response.status_code} "
                f"{' content=' + str(response.content) if self.debug else ''}",
                status=response.status_code,
            )
        return json.loads(response.text)["session_id"]



"""Function strategy."""
from oteapi.models import FunctionConfig

from otelib.backends.services.base import BaseServicesStrategy


class Function(BaseServicesStrategy):
    """Context class for the Function Strategy Interfaces"""

    strategy_name = "function"
    strategy_config = FunctionConfig



"""Configuration settings for creating the OTE client."""
from pydantic import BaseModel, Field


class Settings(BaseModel):
    """Configuration settings for an OTE client."""

    prefix: str = Field("/api/v1", description="Application route prefix.")
    timeout: tuple[float, float] = Field(
        (3.0, 27.0), description="Tuple for URL connect and read timeouts in seconds."
    )

    class Config:
        """Pydantic configuration class."""

        env_prefix = "OTEAPI_"



"""Common strategy for Download, Parse and Resource strategies."""
import json
from copy import deepcopy
from typing import TYPE_CHECKING

from oteapi.models import ResourceConfig
from oteapi.plugins import create_strategy

from otelib.backends.python.base import BasePythonStrategy

if TYPE_CHECKING:  # pragma: no cover
    from typing import Type


class DataResource(BasePythonStrategy):
    """Context class for the data resource strategy interfaces for managing i/o
    operations."""

    strategy_name = "dataresource"
    strategy_config: "Type[ResourceConfig]" = ResourceConfig

    def fetch(self, session_id: str) -> bytes:
        self._sanity_checks(session_id)

        config = self.strategy_config(**json.loads(self.cache[self.strategy_id]))

        if config.downloadUrl and config.mediaType:
            # Download strategy
            session_update = create_strategy("download", config).get(
                session=deepcopy(self.cache[session_id])
            )
            self.cache[session_id].update(session_update)

            # Parse strategy
            session_update = create_strategy("parse", config).get(
                session=deepcopy(self.cache[session_id])
            )
            self.cache[session_id].update(session_update)

        elif config.accessUrl and config.accessService:
            # Resource strategy
            session_update = create_strategy("resource", config).get(
                session=deepcopy(self.cache[session_id])
            )
            self.cache[session_id].update(session_update)

        return session_update.json().encode(encoding="utf-8")

    def initialize(self, session_id: str) -> bytes:
        self._sanity_checks(session_id)

        config = self.strategy_config(**json.loads(self.cache[self.strategy_id]))

        if config.downloadUrl and config.mediaType:
            # Download strategy
            session_update = create_strategy("download", config).initialize(
                session=deepcopy(self.cache[session_id])
            )
            self.cache[session_id].update(session_update)

            # Parse strategy
            session_update = create_strategy("parse", config).initialize(
                session=deepcopy(self.cache[session_id])
            )
            self.cache[session_id].update(session_update)

        elif config.accessUrl and config.accessService:
            # Resource strategy
            session_update = create_strategy("resource", config).initialize(
                session=deepcopy(self.cache[session_id])
            )
            self.cache[session_id].update(session_update)

        return session_update.json().encode(encoding="utf-8")



"""Function strategy."""
from typing import TYPE_CHECKING

from oteapi.models import FunctionConfig

from otelib.backends.python.base import BasePythonStrategy

if TYPE_CHECKING:  # pragma: no cover
    from typing import Type


class Function(BasePythonStrategy):
    """Context class for the function resource strategy interfaces for managing i/o
    operations."""

    strategy_name = "function"
    strategy_config: "Type[FunctionConfig]" = FunctionConfig



"""Filter strategy."""
from typing import TYPE_CHECKING

from oteapi.models import FilterConfig

from otelib.backends.python.base import BasePythonStrategy

if TYPE_CHECKING:  # pragma: no cover
    from typing import Type


class Filter(BasePythonStrategy):
    """Context class for the filter strategy interfaces for managing i/o
    operations."""

    strategy_name = "filter"
    strategy_config: "Type[FilterConfig]" = FilterConfig



"""Client for services backend."""
from typing import TYPE_CHECKING

from otelib.backends.client import AbstractBaseClient

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Dict, Type

    from otelib.backends.services.base import BaseServicesStrategy


# pylint: disable=duplicate-code
class OTEServiceClient(AbstractBaseClient):
    """The Service version of the OTEClient object representing a remote OTE REST API.

    Attributes:
        url (str): The base URL of the OTEAPI Service.

    """

    _backend = "services"

    # config
    _headers: "Dict[str, Any]" = {}

    @property
    def url(self) -> str:
        """Proxy for the source attribute."""
        return self.source

    def _create_strategy(  # type: ignore[override]
        self, strategy_cls: "Type[BaseServicesStrategy]", **config
    ) -> "BaseServicesStrategy":
        strategy = strategy_cls(self.url)
        strategy.headers = self.headers
        strategy.create(**config)
        return strategy

    @property
    def headers(self) -> "Dict[str, Any]":
        """URL headers to use for all requests to the OTEAPI Service."""
        value = self._headers
        if "Content-Type" not in value:
            value["Content-Type"] = "application/json"
        return value

    @headers.setter
    def headers(self, value: "Dict[str, Any]") -> None:
        """Set the URL headers to use for all requests to the OTEAPI Service."""
        if not isinstance(value, dict):
            raise TypeError("headers must be a dictionary")
        self._headers = value

    def _set_config(self, config: "Dict[str, Any]") -> None:
        self.headers = config.pop("headers", {})
        return super()._set_config(config)



"""Test the `otelib.pipe` module."""
# pylint: disable=too-many-arguments,too-many-locals,protected-access,redefined-builtin
# pylint: disable=too-many-branches
from typing import TYPE_CHECKING

import pytest
from utils import strategy_create_kwargs

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Union

    from otelib.backends import python as python_backend
    from otelib.backends import services as services_backend
    from otelib.backends.python.base import BasePythonStrategy
    from otelib.backends.services.base import BaseServicesStrategy

    from .conftest import OTEResponse, ResourceType

    BaseStrategy = Union[BasePythonStrategy, BaseServicesStrategy]

    DataResource = Union[python_backend.DataResource, services_backend.DataResource]
    Filter = Union[python_backend.Filter, services_backend.Filter]


@pytest.mark.parametrize(
    "strategy_name,create_kwargs",
    strategy_create_kwargs(),
    ids=[_[0] for _ in strategy_create_kwargs()],
)
@pytest.mark.usefixtures("mock_session")
def test_pipe(
    backend: str,
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    testdata: "Callable[[Union[ResourceType, str]], dict]",
    server_url: str,
    strategy_name: str,
    create_kwargs: "Dict[str, Any]",
) -> None:
    """Test creating a `Pipe` and run the `get()` method."""
    if strategy_name == "function":
        if backend == "services" and "example" in server_url:
            pass
        else:
            pytest.skip("No function strategy exists in oteapi-core yet.")

    import importlib
    import json

    import requests

    from otelib.pipe import Pipe

    strategies = importlib.import_module(f"otelib.backends.{backend}")
    server_url = server_url if backend != "python" else backend

    if backend == "services":
        # Mock URL responses

        # create()
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_name}",
            return_json={
                f"{strategy_name[len('data'):] if strategy_name.startswith('data') else strategy_name}"  # pylint: disable=line-too-long
                "_id": ids(strategy_name)
            },
        )

        # initialize()
        # The filter and mapping returns everything from their `initialize()` method.
        mock_ote_response(
            method="post",
            endpoint=f"/{strategy_name}/{ids(strategy_name)}/initialize",
            params={"session_id": ids("session")},
            return_json=(
                testdata(strategy_name)
                if strategy_name in ("filter", "mapping")
                else {}
            ),
        )

        # fetch()
        # The data resource and transformation returns everything from their `get()`
        # method.
        mock_ote_response(
            method="get",
            endpoint=f"/{strategy_name}/{ids(strategy_name)}",
            params={"session_id": ids("session")},
            return_json=(
                testdata(strategy_name)
                if strategy_name in ("dataresource", "transformation")
                else {}
            ),
        )

        # Session content
        mock_ote_response(
            method="get",
            endpoint=f"/session/{ids('session')}",
            return_json=testdata(strategy_name),
        )

    strategy_name_map = {"dataresource": "DataResource"}

    strategy: "BaseStrategy" = getattr(
        strategies, strategy_name_map.get(strategy_name, strategy_name.capitalize())
    )(server_url)

    # We must create the strategy - getting a strategy ID
    strategy.create(**create_kwargs)
    assert strategy.strategy_id

    pipe = Pipe(strategy)

    content = pipe.get()
    if strategy_name in ("filter", "mapping"):
        assert json.loads(content) == {}
    elif (
        strategy_name in ("transformation",)
        and backend == "services"
        and "example" not in server_url
    ):
        # Real backend !
        # Dynamic response content - just check keys are the same and values are
        # non-empty
        _content: "dict[str, Any]" = json.loads(content)
        assert list(_content) == list(testdata(strategy_name))
        assert all(_content.values())
    else:
        assert json.loads(content) == testdata(strategy_name)

    # The testdata should always be in the full session
    if backend == "services":
        content_session = requests.get(
            f"{strategy.url}{strategy.settings.prefix}/session/{strategy._session_id}",
            timeout=30,
        )
        session: "Dict[str, Any]" = content_session.json()
    elif backend == "python":
        session_ids = [x for x in strategy.cache if "session" in x]
        assert len(session_ids) == 1
        session_id = session_ids[0]
        session = strategy.cache[session_id]
    for key, value in testdata(strategy_name).items():
        assert key in session

        if strategy_name == "mapping" and key == "triples":
            # The mapping strategy's "triples" key has a Set type value
            session_triples = sorted(list(triple) for triple in session[key])
            assert sorted(value) == session_triples
        elif (
            strategy_name == "transformation"
            and key == "celery_task_id"
            and "example" not in server_url
        ):
            # The task ID is dynamically generated.
            # Simply check the value is non-empty
            assert key in session and session[key]
        else:
            assert value == session[key]


@pytest.mark.usefixtures("mock_session")
def test_pipeing_strategies(  # pylint: disable=too-many-statements
    backend: str,
    mock_ote_response: "OTEResponse",
    ids: "Callable[[Union[ResourceType, str]], str]",
    testdata: "Callable[[Union[ResourceType, str]], dict]",
    server_url: str,
) -> None:
    """A simple pipeline will be tested."""
    import importlib
    import json

    import requests

    strategies = importlib.import_module(f"otelib.backends.{backend}")
    server_url = server_url if backend != "python" else backend

    session_test_content = testdata("filter")
    session_test_content.update(testdata("dataresource"))

    if backend == "services":
        # Mock URL responses

        # create()
        mock_ote_response(
            method="post",
            endpoint="/dataresource",
            return_json={"resource_id": ids("dataresource")},
        )
        mock_ote_response(
            method="post",
            endpoint="/filter",
            return_json={"filter_id": ids("filter")},
        )

        # initialize()
        mock_ote_response(
            method="post",
            endpoint=f"/dataresource/{ids('dataresource')}/initialize",
            params={"session_id": ids("session")},
            return_json={},
        )
        mock_ote_response(
            method="post",
            endpoint=f"/filter/{ids('filter')}/initialize",
            params={"session_id": ids("session")},
            return_json=testdata("filter"),
        )

        # fetch()
        mock_ote_response(
            method="get",
            endpoint=f"/dataresource/{ids('dataresource')}",
            params={"session_id": ids("session")},
            return_json=testdata("dataresource"),
        )
        mock_ote_response(
            method="get",
            endpoint=f"/filter/{ids('filter')}",
            params={"session_id": ids("session")},
            return_json={},
        )

        # Session content
        mock_ote_response(
            method="get",
            endpoint=f"/session/{ids('session')}",
            return_json=session_test_content,
        )

    strategy_kwargs = {}
    if backend == "python":
        # Setup custom cache
        cache = {}
        strategy_kwargs["cache"] = cache

    data_resource: "DataResource" = strategies.DataResource(
        server_url, **strategy_kwargs
    )
    filter: "Filter" = strategies.Filter(server_url, **strategy_kwargs)

    # We must create the data resource and filter - getting IDs
    create_kwargs = dict(strategy_create_kwargs())
    data_resource.create(**create_kwargs["dataresource"])
    assert data_resource.strategy_id
    filter.create(**create_kwargs["filter"])
    assert filter.strategy_id

    pipeline = data_resource >> filter

    content = pipeline.get()

    # Since the "last" strategy does not return anything from its `get()` method,
    # this should return an empty dictionary.
    assert json.loads(content) == {}

    # All the test data should however be present in the session
    assert (
        pipeline._session_id
    ), f"Session ID not found in {pipeline} ! Is OTEAPI_DEBUG not set?"

    if backend == "services":
        content_session = requests.get(
            f"{pipeline.url}{pipeline.settings.prefix}/session/{pipeline._session_id}",
            timeout=30,
        )
        session: "Dict[str, Any]" = content_session.json()
    elif backend == "python":
        session_ids = [x for x in pipeline.cache if "session" in x]
        assert len(session_ids) == 1
        session_id = session_ids[0]
        session = pipeline.cache[session_id]
    for key, value in session_test_content.items():
        assert key in session
        assert value == session[key]

    ##
    # Reverse the pipeline and try again
    ##
    strategy_kwargs = {}
    if backend == "python":
        # Setup custom cache
        cache = {}
        strategy_kwargs["cache"] = cache

    data_resource: "DataResource" = strategies.DataResource(
        server_url, **strategy_kwargs
    )
    filter: "Filter" = strategies.Filter(server_url, **strategy_kwargs)

    # We must create the data resource and filter - getting IDs
    create_kwargs = dict(strategy_create_kwargs())
    data_resource.create(**create_kwargs["dataresource"])
    assert data_resource.strategy_id
    filter.create(**create_kwargs["filter"])
    assert filter.strategy_id

    pipeline = filter >> data_resource

    content = pipeline.get()

    # Since the "last" strategy now returns something from its `get()` method,
    # this can be tested.
    assert json.loads(content) == testdata("dataresource")

    # All the test data should however be present in the session
    assert (
        pipeline._session_id
    ), f"Session ID not found in {pipeline} ! Is OTEAPI_DEBUG not set?"

    if backend == "services":
        content_session = requests.get(
            f"{pipeline.url}{pipeline.settings.prefix}/session/{pipeline._session_id}",
            timeout=30,
        )
        session: "Dict[str, Any]" = content_session.json()
    elif backend == "python":
        session_ids = [x for x in pipeline.cache if "session" in x]
        assert len(session_ids) == 1
        session_id = session_ids[0]
        session = pipeline.cache[session_id]
    for key, value in session_test_content.items():
        assert key in session
        assert value == session[key]


def test_pipeing_concatenate(  # pylint: disable=too-many-statements
    backend: str,
    server_url: str,
) -> None:
    """Tests for the correct chaining of resources and filters in a pipeline.
    This function tests for issue 110: https://github.com/EMMC-ASBL/otelib/issues/110
    """

    import importlib

    strategies = importlib.import_module(f"otelib.backends.{backend}")
    server_url = server_url if backend != "python" else backend

    strategy_kwargs = {}
    if backend == "python":
        # Setup custom cache
        cache = {}
        strategy_kwargs["cache"] = cache

    data_resource1: "DataResource" = strategies.DataResource(
        server_url, **strategy_kwargs
    )
    dataid1 = data_resource1.strategy_id
    filter1: "Filter" = strategies.Filter(server_url, **strategy_kwargs)
    filterid1 = filter1.strategy_id
    pipeline1 = data_resource1 >> filter1

    data_resource2: "DataResource" = strategies.DataResource(
        server_url, **strategy_kwargs
    )
    dataid2 = data_resource2.strategy_id
    filter2: "Filter" = strategies.Filter(server_url, **strategy_kwargs)
    filterid2 = filter2.strategy_id
    pipeline2 = data_resource2 >> filter2

    pipeline_combined = pipeline1 >> pipeline2

    # confirm that we can retrieve the strategy ids
    pipeline_tail = pipeline_combined  # or should it be pipeline head?
    assert pipeline_tail.strategy_id == dataid1
    pipeline_tail = pipeline_tail.input_pipe.input
    assert pipeline_tail.strategy_id == filterid1
    pipeline_tail = pipeline_tail.input_pipe.input
    assert pipeline_tail.strategy_id == dataid2
    pipeline_tail = pipeline_tail.input_pipe.input
    assert pipeline_tail.strategy_id == filterid2



"""Mapping strategy."""
from oteapi.models import MappingConfig

from otelib.backends.services.base import BaseServicesStrategy


class Mapping(BaseServicesStrategy):
    """Context class for the Mapping Strategy Interfaces"""

    strategy_name = "mapping"
    strategy_config = MappingConfig



"""Common strategy for Download, Prase and Resource strategies."""
from oteapi.models import ResourceConfig

from otelib.backends.services.base import BaseServicesStrategy


class DataResource(BaseServicesStrategy):
    """Context class for the data resource strategy interfaces for managing i/o
    operations."""

    strategy_name = "dataresource"
    strategy_config = ResourceConfig
